/*
 * CNsort.c - procedure declarations for sorting
 */

#ifndef CNsort_defined
#define CNsort_defined

extern void   CNdo_quick_sort_curves();
extern void   CNbubble_sort_curves();

extern void   CNdo_quick_sort_trias();
extern void   CNbubble_sort_trias();

extern void   CNdo_quick_sort_rects();
extern void   CNbubble_sort_rects();

extern void   CNdo_quick_sort_cubes();
extern void   CNbubble_sort_cubes();

extern void   CNdo_quick_sort_blocks();
extern void   CNbubble_sort_blocks();

extern void   CNdo_quick_sort_prisms();
extern void   CNbubble_sort_prisms();

extern void   CNdo_quick_sort_polys();
extern void   CNbubble_sort_polys();

extern void   CNdo_quick_sort_xpoints();

#endif /* CNsort_defined */

