/*
 * CNmesh.h - definitions for a triangular mesh
 */

#ifndef CNmesh_defined
#define CNmesh_defined

#define CN_MAXOBJS  30000
#define CN_MAXSOLS  60
#define CN_MESHNAME "Mesh Structure"

#endif /* CNmesh_defined */

